import functools
import inspect
import json
import time
from collections import defaultdict
from typing import (
    Any,
    Awaitable,
    Callable,
    Hashable,
    Literal,
    Mapping,
    TypeAlias,
    TypeVar,
)

import yaml
from aiohttp import web, web_response
from aiohttp.typedefs import Handler
from pydantic import BaseModel, ValidationError

from ai.backend.logging import BraceStyleAdapter
from ai.backend.wsproxy.defs import RootContext
from ai.backend.wsproxy.exceptions import AuthorizationFailed

from ..exceptions import InvalidAPIParameters
from ..types import PydanticResponse


def auth_required(scope: Literal["manager"] | Literal["worker"]) -> Callable[[Handler], Handler]:
    def wrap(handler: Handler) -> Handler:
        @functools.wraps(handler)
        async def wrapped(request: web.Request, *args, **kwargs):
            root_ctx: RootContext = request.app["_root.context"]
            permitted_token = root_ctx.local_config.wsproxy.api_secret
            permitted_header_values = (
                permitted_token,
                f"Bearer {permitted_token}",
                f"BackendAI {permitted_token}",
            )
            token_to_evaluate = request.headers.get("X-BackendAI-Token")
            if token_to_evaluate not in permitted_header_values:
                raise AuthorizationFailed("Unauthorized access")
            return await handler(request, *args, **kwargs)

        original_attrs = getattr(handler, "_backend_attrs", {})
        for k, v in original_attrs.items():
            set_handler_attr(wrapped, k, v)

        set_handler_attr(wrapped, "auth_scope", scope)
        return wrapped

    return wrap


# FIXME: merge majority of common definitions to ai.backend.common when ready

_danger_words = ["password", "passwd", "secret"]


def set_handler_attr(func, key, value):
    attrs = getattr(func, "_backend_attrs", None)
    if attrs is None:
        attrs = {}
    attrs[key] = value
    setattr(func, "_backend_attrs", attrs)


def get_handler_attr(request, key, default=None):
    # When used in the aiohttp server-side codes, we should use
    # request.match_info.hanlder instead of handler passed to the middleware
    # functions because aiohttp wraps this original handler with functools.partial
    # multiple times to implement its internal middleware processing.
    attrs = getattr(request.match_info.handler, "_backend_attrs", None)
    if attrs is not None:
        return attrs.get(key, default)
    return default


def mask_sensitive_keys(data: Mapping[str, Any]) -> Mapping[str, Any]:
    """
    Returns a new cloned mapping by masking the values of
    sensitive keys with "***" from the given mapping.
    """
    sanitized = dict()
    for k, v in data.items():
        if any((w in k.lower()) for w in _danger_words):
            sanitized[k] = "***"
        else:
            sanitized[k] = v
    return sanitized


_burst_last_call: float = 0.0
_burst_times: dict[Hashable, float] = dict()
_burst_counts: dict[Hashable, int] = defaultdict(int)


async def call_non_bursty(
    key: Hashable,
    coro: Callable[[], Any],
    *,
    max_bursts: int = 64,
    max_idle: int | float = 100.0,
):
    """
    Execute a coroutine once upon max_bursts bursty invocations or max_idle
    milliseconds after bursts smaller than max_bursts.
    """
    global _burst_last_call, _burst_times, _burst_counts
    if inspect.iscoroutine(coro):
        # Coroutine objects may not be called before garbage-collected
        # as this function throttles the frequency of invocation.
        # That will generate a bogus warning by the asyncio's debug facility.
        raise TypeError("You must pass coroutine function, not coroutine object.")
    now = time.monotonic()

    if now - _burst_last_call > 3.0:
        # garbage-collect keys
        cleaned_keys = []
        for k, tick in _burst_times.items():
            if now - tick > (max_idle / 1e3):
                cleaned_keys.append(k)
        for k in cleaned_keys:
            del _burst_times[k]
            _burst_counts.pop(k, None)

    last_called = _burst_times.get(key, 0)
    _burst_times[key] = now
    _burst_last_call = now
    invoke = False

    if now - last_called > (max_idle / 1e3):
        invoke = True
        _burst_counts.pop(key, None)
    else:
        _burst_counts[key] += 1
    if _burst_counts[key] >= max_bursts:
        invoke = True
        del _burst_counts[key]

    if invoke:
        if inspect.iscoroutinefunction(coro):
            return await coro()
        else:
            return coro()


TAnyResponse = TypeVar("TAnyResponse", bound=web.StreamResponse)

TParamModel = TypeVar("TParamModel", bound=BaseModel)
TQueryModel = TypeVar("TQueryModel", bound=BaseModel)
TResponseModel = TypeVar("TResponseModel", bound=BaseModel)

THandlerFuncWithoutParam: TypeAlias = Callable[
    [web.Request], Awaitable[PydanticResponse | TAnyResponse]
]
THandlerFuncWithParam: TypeAlias = Callable[
    [web.Request, TParamModel], Awaitable[PydanticResponse | TAnyResponse]
]


def ensure_stream_response_type(
    response: PydanticResponse | TAnyResponse,
) -> web.StreamResponse:
    json_body: Any
    match response:
        case PydanticResponse():
            match response.response:
                case BaseModel():
                    json_body = response.response.model_dump(mode="json")
                case _:
                    raise RuntimeError(f"Unsupported model type ({type(response.response)})")
            return web.json_response(json_body, headers=response.headers, status=response.status)
        case web_response.StreamResponse():
            return response
        case _:
            raise RuntimeError(f"Unsupported response type ({type(response)})")


def pydantic_api_response_handler(
    handler: THandlerFuncWithoutParam,
    is_deprecated=False,
) -> Handler:
    """
    Only for API handlers which does not require request body.
    For handlers with params to consume use @pydantic_params_api_handler() or
    @check_api_params() decorator (only when request param is validated with trafaret).
    """

    @functools.wraps(handler)
    async def wrapped(
        request: web.Request,
        *args,
        **kwargs,
    ) -> web.StreamResponse:
        response = await handler(request, *args, **kwargs)
        return ensure_stream_response_type(response)

    set_handler_attr(wrapped, "deprecated", is_deprecated)
    return wrapped


def pydantic_api_handler(
    checker: type[TParamModel],
    loads: Callable[[str], Any] | None = None,
    query_param_checker: type[TQueryModel] | None = None,
    is_deprecated=False,
) -> Callable[[THandlerFuncWithParam], Handler]:
    def wrap(
        handler: THandlerFuncWithParam,
    ) -> Handler:
        @functools.wraps(handler)
        async def wrapped(
            request: web.Request,
            *args,
            **kwargs,
        ) -> web.StreamResponse:
            orig_params: Any
            body: str = ""
            log: BraceStyleAdapter = request["log"]
            try:
                body_exists = request.can_read_body
                if body_exists:
                    body = await request.text()
                    if request.content_type == "text/yaml":
                        orig_params = yaml.load(body, Loader=yaml.BaseLoader)
                    else:
                        orig_params = (loads or json.loads)(body)
                else:
                    orig_params = dict(request.query)
                stripped_params = orig_params.copy()
                log.debug("stripped raw params: {}", mask_sensitive_keys(stripped_params))
                checked_params = checker.model_validate(stripped_params)
                if body_exists and query_param_checker:
                    query_params = query_param_checker.model_validate(request.query)
                    kwargs["query"] = query_params
            except (json.decoder.JSONDecodeError, yaml.YAMLError, yaml.MarkedYAMLError):
                raise InvalidAPIParameters("Malformed body")
            except ValidationError as e:
                raise InvalidAPIParameters("Input validation error", extra_data=e.errors())
            result = await handler(request, checked_params, *args, **kwargs)
            return ensure_stream_response_type(result)

        original_attrs = getattr(handler, "_backend_attrs", {})
        for k, v in original_attrs.items():
            set_handler_attr(wrapped, k, v)

        set_handler_attr(wrapped, "request_scheme", checker)
        set_handler_attr(wrapped, "deprecated", is_deprecated)

        return wrapped

    return wrap



import sys
import textwrap
from collections import defaultdict
from typing import Any, Union, get_args, get_origin, get_type_hints

from aiohttp import web
from aiohttp.web_urldispatcher import AbstractResource, DynamicResource
from pydantic import BaseModel

from . import __version__
from .types import PydanticResponse


class ParseError(Exception):
    pass


def get_path_parameters(resource: AbstractResource) -> list[dict]:
    params = []
    if isinstance(resource, DynamicResource):
        if groupindex := resource._pattern.groupindex:
            params = [
                {"name": param, "in": "path", "required": True, "schema": {"type": "string"}}
                for param in groupindex.keys()
            ]
    return params


def generate_openapi(
    component: str, subapps: list[web.Application], verbose=False
) -> dict[str, Any]:
    openapi: dict[str, Any] = {
        "openapi": "3.1.0",
        "info": {
            "title": f"Backend.AI {component} API",
            "description": f"Backend.AI {component} REST API specification",
            "version": __version__,
            "contact": {
                "name": "Lablup Inc.",
                "url": "https://docs.backend.ai",
                "email": "contect@lablup.com",
            },
        },
        "components": {
            "securitySchemes": {
                "X-BackendAI-Token": {
                    "type": "http",
                    "scheme": "bearer",
                    "bearerFormat": "JWT",
                },
            },
            "schemas": {},
        },
        "paths": defaultdict(lambda: {}),
    }
    operation_id_mapping: defaultdict[str, int] = defaultdict(lambda: 0)
    for app in subapps:
        prefix = app.get("prefix", "root")
        for route in app.router.routes():
            resource = route.resource
            if not resource:
                continue

            if "_root_app" not in app:
                path = "/" + ("" if prefix == "root" else prefix) + resource.canonical
            else:
                path = resource.canonical
            method = route.method

            if method == "OPTIONS":
                continue

            operation_id = f"{prefix}.{route.handler.__name__}"
            if verbose:
                sys.stderr.write(f"parsing {operation_id}\n")
                sys.stderr.flush()
            operation_id_mapping[operation_id] += 1
            if (operation_id_count := operation_id_mapping[operation_id]) > 1:
                operation_id += f".{operation_id_count}"

            description = []
            if route.handler.__doc__:
                description.append(textwrap.dedent(route.handler.__doc__))

            route_def = {
                "operationId": operation_id,
                "tags": [prefix],
                "responses": dict(),
            }
            parameters = []
            parameters.extend(get_path_parameters(resource))
            if hasattr(route.handler, "_backend_attrs"):
                preconds = []
                handler_attrs = getattr(route.handler, "_backend_attrs")
                if handler_attrs.get("auth_required"):
                    route_def["security"] = [{"X-BackendAI-Token": []}]
                if auth_scope := handler_attrs.get("auth_scope"):
                    preconds.append(
                        f"Requires {auth_scope.capitalize()} token present at `X-BackendAI-Token` request header to work."
                    )
                if preconds:
                    description.append("\n**Preconditions:**")
                    for item in preconds:
                        description.append(f"* {item}")
                    description.append("")
                if request_scheme := handler_attrs.get("request_scheme"):
                    if issubclass(request_scheme, BaseModel):
                        schema_name = request_scheme.__name__
                        request_schema = request_scheme.model_json_schema(
                            ref_template="#/components/schemas/{model}"
                        )

                        if additional_definitions := request_schema.pop("$defs", None):
                            openapi["components"]["schemas"].update(additional_definitions)
                        openapi["components"]["schemas"][schema_name] = request_schema
                        route_def["requestBody"] = {
                            "deprecated": handler_attrs.get("deprecated", False),
                            "content": {
                                "application/json": {
                                    "schema": {"$ref": f"#/components/schemas/{schema_name}"}
                                }
                            },
                        }
                    else:
                        raise RuntimeError(
                            f"{request_scheme} not considered as a valid request type"
                        )

            route_def["parameters"] = parameters
            route_def["description"] = "\n".join(description)
            type_hints = get_type_hints(route.handler)

            def _parse_schema(model_cls: type[BaseModel]) -> dict:
                if not issubclass(model_cls, BaseModel):
                    raise RuntimeError(f"{model_cls} not considered as a valid response type")

                schema_name = model_cls.__name__
                response_schema = model_cls.model_json_schema(
                    ref_template="#/components/schemas/{model}"
                )

                if additional_definitions := response_schema.pop("$defs", None):
                    openapi["components"]["schemas"].update(additional_definitions)
                openapi["components"]["schemas"][schema_name] = response_schema
                return {
                    "description": "",
                    "content": {
                        "application/json": {
                            "schema": {"$ref": f"#/components/schemas/{schema_name}"}
                        }
                    },
                }

            if (ret_type := type_hints.get("return")) and (ret_type_origin := get_origin(ret_type)):
                if ret_type_origin == Union:
                    response_classes = get_args(ret_type)
                    responses = dict()

                    for cls in response_classes:
                        if (subclass_origin := get_origin(cls)) and issubclass(
                            subclass_origin, PydanticResponse
                        ):
                            if "200" in responses:
                                raise RuntimeError(
                                    "Cannot specify multiple response types for a single API handler"
                                )
                            responses["200"] = _parse_schema(get_args(cls)[0])
                        elif issubclass(cls, web.HTTPTemporaryRedirect):
                            responses["301"] = {"Description": "Redirection"}
                        elif issubclass(cls, web.HTTPPermanentRedirect):
                            responses["302"] = {"Description": "Redirection"}

                    route_def["responses"] = responses
                elif issubclass(ret_type_origin, PydanticResponse):
                    route_def["responses"] = {"200": _parse_schema(get_args(ret_type)[0])}

            openapi["paths"][path][method.lower()] = route_def
    return openapi



from typing import TYPE_CHECKING, AsyncContextManager, Callable, TypeAlias
from uuid import UUID

import aiohttp_cors
import attrs

from .config import ServerConfig

if TYPE_CHECKING:
    from .proxy.frontend.abc import AbstractFrontend


@attrs.define(slots=True, auto_attribs=True, init=False)
class RootContext:
    pidx: int
    proxy_frontend: "AbstractFrontend"
    worker_id: UUID
    local_config: ServerConfig
    cors_options: dict[str, aiohttp_cors.ResourceOptions]


CleanupContext: TypeAlias = Callable[["RootContext"], AsyncContextManager[None]]
